@@ -10,12 +10,20 @@ int nl_init_event_handler(void);
 
 int nl_init_dump_handler(void);
 
+int nl_init_overrun_handler(void);
+
+int nl_overrun_request_resync(void);
+
 void nl_resize_socket_buffer(struct nfct_handle *h);
 
 int nl_dump_conntrack_table(void);
 
+int nl_exist_conntrack(struct nf_conntrack *ct);
+
 int nl_create_conntrack(struct nf_conntrack *ct);
 
+int nl_update_conntrack(struct nf_conntrack *ct);
+
 int nl_destroy_conntrack(struct nf_conntrack *ct);
 
 #endif
