@@ -98,10 +98,14 @@
      T_SYSLOG = 314,
      T_WRITE_THROUGH = 315,
      T_STAT_BUFFER_SIZE = 316,
-     T_IP = 317,
-     T_PATH_VAL = 318,
-     T_NUMBER = 319,
-     T_STRING = 320
+     T_DESTROY_TIMEOUT = 317,
+     T_MCAST_RCVBUFF = 318,
+     T_MCAST_SNDBUFF = 319,
+     T_NOTRACK = 320,
+     T_IP = 321,
+     T_PATH_VAL = 322,
+     T_NUMBER = 323,
+     T_STRING = 324
    };
 #endif
 /* Tokens.  */
@@ -164,10 +168,14 @@
 #define T_SYSLOG 314
 #define T_WRITE_THROUGH 315
 #define T_STAT_BUFFER_SIZE 316
-#define T_IP 317
-#define T_PATH_VAL 318
-#define T_NUMBER 319
-#define T_STRING 320
+#define T_DESTROY_TIMEOUT 317
+#define T_MCAST_RCVBUFF 318
+#define T_MCAST_SNDBUFF 319
+#define T_NOTRACK 320
+#define T_IP 321
+#define T_PATH_VAL 322
+#define T_NUMBER 323
+#define T_STRING 324
 
 
 
@@ -180,7 +188,7 @@ typedef union YYSTYPE
 	char		*string;
 }
 /* Line 1489 of yacc.c.  */
-#line 184 "read_config_yy.h"
+#line 192 "read_config_yy.h"
 	YYSTYPE;
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
